from .Subfinder import *
from .DanbooruFinder import DanbooruFinder
from .Rule34Finder import Rule34Finder
from .KonachanFinder import KonachanFinder
from .YandereFinder import YandereFinder
from .GelbooruFinder import GelbooruFinder
from .PixivFinder import PixivFinder
from .ManganatoFinder import ManganatoFinder
from .NHentaiFinder import NHentaiFinder

builtin_finders = [
    DanbooruFinder,
    Rule34Finder,
    KonachanFinder,
    YandereFinder,
    GelbooruFinder,
    PixivFinder,
    ManganatoFinder
]